==> imate-0.5.3-cp36-cp36m-macosx_10_9_x86_64.whl: imate/traceinv/_eigenvalue_method.py <==
# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the license found in the LICENSE.txt file in the root
# directory of this source tree.


# =======
# Imports
# =======

import time
import numpy
import numpy.linalg
import scipy
import scipy.linalg
import scipy.sparse
import scipy.sparse.linalg


# =================
# eigenvalue method
# =================

def eigenvalue_method(
        A,
        eigenvalues=None,
        exponent=1.0,
        symmetric=True,
        non_zero_eig_fraction=0.9):
    """
    """

    # Checking input arguments
    check_arguments(A, eigenvalues, exponent, symmetric,
                    non_zero_eig_fraction)

    init_wall_time = time.perf_counter()
    init_proc_time = time.process_time()

    if eigenvalues is None:

        # This is only relevant when A is sparse and we can only find a
        # portion, bit not all, of its eigenvalues.
        if exponent >= 0:
            # Since we compute the trace of inverse, when the exponent is
            # positive, the smallest eigenvalues weight the most in the trace
            # of inverse.
            which_eigenvalues = 'SM'
        else:
            # In contrary, when the exponent is negative, the largest
            # eigenvalues weight more inthe trace of inverse.
            which_eigenvalues = 'LM'

        # Compute eigenvalues of A
        eigenvalues = compute_eigenvalues(A, symmetric,
                                          non_zero_eig_fraction,
                                          which_eigenvalues)

    # Compute trace of inverse of matrix
    not_nan = numpy.logical_not(numpy.isnan(eigenvalues))
    trace = numpy.sum(1.0 / (eigenvalues[not_nan]**exponent))

    wall_time = time.perf_counter() - init_wall_time
    proc_time = time.process_time() - init_proc_time

    # Dictionary of output info
    info = {
        'cpu':
        {
            'wall_time': wall_time,
            'proc_time': proc_time,
        },
        'solver':
        {
            'method': 'eigenvalue',
        }
    }

    return trace, info


# ===============
# check arguments
# ===============

def check_arguments(
        A,
        eigenvalues,
        exponent,
        symmetric,
        non_zero_eig_fraction):
    """
    Checks the type and value of the parameters.
    """

    # Check A
    if (not isinstance(A, numpy.ndarray)) and (not scipy.sparse.issparse(A)):
        raise TypeError('Input matrix should be either a "numpy.ndarray" or ' +
                        'a "scipy.sparse" matrix.')
    elif A.shape[0] != A.shape[1]:
        raise ValueError('Input matrix should be a square matrix.')

    # Check eigenvalues
    if eigenvalues is not None:
        if not isinstance(eigenvalues, numpy.ndarray):
            raise TypeError('"eigenvalues" should be a numpy.ndarray.')
        if eigenvalues.size != A.shape[0]:
            raise ValueError('The length of "eigenvalues" does not match ' +
                             'the size of matrix "A".')

    # Check exponent
    if exponent is None:
        raise TypeError('"exponent" cannot be None.')
    elif not numpy.isscalar(exponent):
        raise TypeError('"exponent" should be a scalar value.')
    elif not isinstance(exponent, int):
        TypeError('"exponent" cannot be an integer.')

    # Check symmetric
    if symmetric is None:
        raise TypeError('"symmetric" cannot be None.')
    elif not numpy.isscalar(symmetric):
        raise TypeError('"symmetric" should be a scalar value.')
    elif not isinstance(symmetric, bool):
        raise TypeError('"symmetric" should be boolean.')

    # Check non_zero_eig_fraction
    if non_zero_eig_fraction is None:
        raise TypeError('"non_zero_eig_fraction" cannot be None.')
    elif not numpy.isscalar(non_zero_eig_fraction):
        raise TypeError('"non_zero_eig_fraction" should be a scalar value.')
    elif not isinstance(non_zero_eig_fraction, float):
        raise TypeError('"non_zero_eig_fraction" should be a float type.')
    elif non_zero_eig_fraction <= 0 or non_zero_eig_fraction >= 1.0:
        raise ValueError('"non_zero_eig_fraction" should be greater then 0.0' +
                         'and smaller than 1.0.')


# ===================
# compute eigenvalues
# ===================

def compute_eigenvalues(
        A,
        symmetric,
        non_zero_eig_fraction,
        which_eigenvalues,
        tol=1e-4):
    """
    """

    if scipy.sparse.isspmatrix(A):

        # Sparse matrix
        n = A.shape[0]
        eigenvalues = numpy.empty(n)
        eigenvalues[:] = numpy.nan

        # find 90% of eigenvalues, assume the rest are very close to zero.
        num_none_zero_eig = int(n*non_zero_eig_fraction)

        if symmetric:
            eigenvalues[:num_none_zero_eig] = \
                scipy.sparse.linalg.eigsh(A, num_none_zero_eig,
                                          which=which_eigenvalues,
                                          return_eigenvectors=False,
                                          tol=tol)
        else:
            eigenvalues[:num_none_zero_eig] = \
                scipy.sparse.linalg.eigs(A, num_none_zero_eig,
                                         which=which_eigenvalues,
                                         return_eigenvectors=False,
                                         tol=tol)
    else:

        # Dense matrix
        if symmetric:
            eigenvalues = scipy.linalg.eigh(A, check_finite=False,
                                            eigvals_only=True)
        else:
            eigenvalues = scipy.linalg.eig(A, check_finite=False)[0]

    return eigenvalues

==> imate-0.6.1-cp36-cp36m-macosx_10_9_x86_64.whl: imate/traceinv/_eigenvalue_method.py <==
# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the license found in the LICENSE.txt file in the root
# directory of this source tree.


# =======
# Imports
# =======

import time
import numpy
import numpy.linalg
import scipy
import scipy.linalg
import scipy.sparse
import scipy.sparse.linalg
from scipy.sparse import isspmatrix
import multiprocessing
from .._linear_algebra.matrix_utilities import get_data_type_name, get_nnz, \
        get_density
from ..__version__ import __version__


# =================
# eigenvalue method
# =================

def eigenvalue_method(
        A,
        eigenvalues=None,
        exponent=1.0,
        symmetric=True,
        non_zero_eig_fraction=0.9):
    """
    """

    # Checking input arguments
    check_arguments(A, eigenvalues, exponent, symmetric,
                    non_zero_eig_fraction)

    init_tot_wall_time = time.perf_counter()
    init_cpu_proc_time = time.process_time()

    if eigenvalues is None:

        # This is only relevant when A is sparse and we can only find a
        # portion, bit not all, of its eigenvalues.
        if exponent >= 0:
            # Since we compute the trace of inverse, when the exponent is
            # positive, the smallest eigenvalues weight the most in the trace
            # of inverse.
            which_eigenvalues = 'SM'
        else:
            # In contrary, when the exponent is negative, the largest
            # eigenvalues weight more inthe trace of inverse.
            which_eigenvalues = 'LM'

        # Compute eigenvalues of A
        eigenvalues = compute_eigenvalues(A, symmetric,
                                          non_zero_eig_fraction,
                                          which_eigenvalues)

    # Compute trace of inverse of matrix
    not_nan = numpy.logical_not(numpy.isnan(eigenvalues))
    trace = numpy.sum(1.0 / (eigenvalues[not_nan]**exponent))

    tot_wall_time = time.perf_counter() - init_tot_wall_time
    cpu_proc_time = time.process_time() - init_cpu_proc_time

    # Dictionary of output info
    info = {
        'matrix':
        {
            'data_type': get_data_type_name(A),
            'exponent': exponent,
            'size': A.shape[0],
            'sparse': isspmatrix(A),
            'nnz': get_nnz(A),
            'density': get_density(A),
            'num_inquiries': 1
        },
        'device':
        {
            'num_cpu_threads': multiprocessing.cpu_count()
        },
        'time':
        {
            'tot_wall_time': tot_wall_time,
            'alg_wall_time': tot_wall_time,
            'cpu_proc_time': cpu_proc_time,
        },
        'solver':
        {
            'version': __version__,
            'method': 'eigenvalue',
        }
    }

    return trace, info


# ===============
# check arguments
# ===============

def check_arguments(
        A,
        eigenvalues,
        exponent,
        symmetric,
        non_zero_eig_fraction):
    """
    Checks the type and value of the parameters.
    """

    # Check A
    if (not isinstance(A, numpy.ndarray)) and (not scipy.sparse.issparse(A)):
        raise TypeError('Input matrix should be either a "numpy.ndarray" or ' +
                        'a "scipy.sparse" matrix.')
    elif A.shape[0] != A.shape[1]:
        raise ValueError('Input matrix should be a square matrix.')

    # Check eigenvalues
    if eigenvalues is not None:
        if not isinstance(eigenvalues, numpy.ndarray):
            raise TypeError('"eigenvalues" should be a numpy.ndarray.')
        if eigenvalues.size != A.shape[0]:
            raise ValueError('The length of "eigenvalues" does not match ' +
                             'the size of matrix "A".')

    # Check exponent
    if exponent is None:
        raise TypeError('"exponent" cannot be None.')
    elif not numpy.isscalar(exponent):
        raise TypeError('"exponent" should be a scalar value.')
    elif not isinstance(exponent, (int, numpy.integer)):
        TypeError('"exponent" cannot be an integer.')

    # Check symmetric
    if symmetric is None:
        raise TypeError('"symmetric" cannot be None.')
    elif not numpy.isscalar(symmetric):
        raise TypeError('"symmetric" should be a scalar value.')
    elif not isinstance(symmetric, bool):
        raise TypeError('"symmetric" should be boolean.')

    # Check non_zero_eig_fraction
    if non_zero_eig_fraction is None:
        raise TypeError('"non_zero_eig_fraction" cannot be None.')
    elif not numpy.isscalar(non_zero_eig_fraction):
        raise TypeError('"non_zero_eig_fraction" should be a scalar value.')
    elif not isinstance(non_zero_eig_fraction, float):
        raise TypeError('"non_zero_eig_fraction" should be a float type.')
    elif non_zero_eig_fraction <= 0 or non_zero_eig_fraction >= 1.0:
        raise ValueError('"non_zero_eig_fraction" should be greater then 0.0' +
                         'and smaller than 1.0.')


# ===================
# compute eigenvalues
# ===================

def compute_eigenvalues(
        A,
        symmetric,
        non_zero_eig_fraction,
        which_eigenvalues,
        tol=1e-4):
    """
    """

    if scipy.sparse.isspmatrix(A):

        # Sparse matrix
        n = A.shape[0]
        eigenvalues = numpy.empty(n)
        eigenvalues[:] = numpy.nan

        # find 90% of eigenvalues, assume the rest are very close to zero.
        num_none_zero_eig = int(n*non_zero_eig_fraction)

        if symmetric:
            eigenvalues[:num_none_zero_eig] = \
                scipy.sparse.linalg.eigsh(A, num_none_zero_eig,
                                          which=which_eigenvalues,
                                          return_eigenvectors=False,
                                          tol=tol)
        else:
            eigenvalues[:num_none_zero_eig] = \
                scipy.sparse.linalg.eigs(A, num_none_zero_eig,
                                         which=which_eigenvalues,
                                         return_eigenvectors=False,
                                         tol=tol)
    else:

        # Dense matrix
        if symmetric:
            eigenvalues = scipy.linalg.eigh(A, check_finite=False,
                                            eigvals_only=True)
        else:
            eigenvalues = scipy.linalg.eig(A, check_finite=False)[0]

    return eigenvalues
